@@ -5,6 +5,12 @@ from stm32loader.main import main as stm32loader_main
 import tempfile
 import requests
 import hashlib
+from packaging.version import parse as parse_version
+
+class UnsupportedFirmware(Exception):
+    pass
+class UnsupportedHardware(Exception):
+    pass
 
 class Controller():
     _HEADER = 0x55
@@ -22,20 +28,26 @@ class Controller():
         self.__ph = serial.Serial(port=portname, baudrate=baudrate, timeout=0.1)
         self.__serial_settings = self.__ph.get_settings()
         self.__fw_file = None
+        self.board_info = self.get_board_info()
 
     def __del__(self):
-        if self.__ph.isOpen():
-            self.__ph.flush()
-            self.__ph.flushInput()
-            self.__ph.flushOutput()
-            self.__ph.close()
+        try:
+            if self.__ph.isOpen():
+                self.__ph.flush()
+                self.__ph.flushInput()
+                self.__ph.flushOutput()
+                self.__ph.close()
+        except AttributeError:
+            pass
+        except Exception as e:
+            raise e
 
     def _writebus(self, data):
         self.__ph.write(data)
 
     def _readbus(self, byte_count):
         data = self.__ph.read(byte_count)
-        if len(data) > 0:
+        if len(data) == byte_count:
             if data[0] == self.__class__._HEADER:
                 if self._crc32(data[:-4]) == data[-4:]:
                     return data
@@ -187,9 +199,9 @@ class OneDOF(Controller):
 
     def set_speed(self, speed):
         if speed != 0:
-            self.__speed = speed if abs(speed) <= self.__class__._MAX_SPEED_ABS else self.__class__._MAX_SPEED_ABS * (speed / abs(speed))
+            self.__speed = int(speed if abs(speed) <= self.__class__._MAX_SPEED_ABS else self.__class__._MAX_SPEED_ABS * (speed / abs(speed)))
         else:
-            self.__speed = speed
+            self.__speed = int(speed)
 
     def enable(self, en):
         self.__config = (self.__config & ~self.__class__._EN_MASK) | (en & self.__class__._EN_MASK)
@@ -231,9 +243,9 @@ class BallBeam(Controller):
 
     def set_servo(self, servo):
         if servo != 0:
-            self.__servo = servo if abs(servo) <= self.__class__._MAX_SERVO_ABS else self.__class__._MAX_SERVO_ABS * (servo / abs(servo))
+            self.__servo = int(servo if abs(servo) <= self.__class__._MAX_SERVO_ABS else self.__class__._MAX_SERVO_ABS * (servo / abs(servo)))
         else:
-            self.__servo = servo
+            self.__servo = int(servo)
 
     def _write(self):
         data = struct.pack("<BBh", self.__class__._HEADER, self.__class__._DEVID, self.__servo)
@@ -263,14 +275,14 @@ class BallBalancingTable(Controller):
 
     def set_servo(self, x, y):
         if x != 0:
-            self.__servo[0] = x if abs(x) <= self.__class__._MAX_SERVO_ABS else self.__class__._MAX_SERVO_ABS * (x / abs(x))
+            self.__servo[0] = int(x if abs(x) <= self.__class__._MAX_SERVO_ABS else self.__class__._MAX_SERVO_ABS * (x / abs(x)))
         else:
-            self.__servo[0] = x
+            self.__servo[0] = int(x)
 
         if y != 0:
-            self.__servo[1] = y if abs(x) <= self.__class__._MAX_SERVO_ABS else self.__class__._MAX_SERVO_ABS * (y / abs(y))
+            self.__servo[1] = int(y if abs(x) <= self.__class__._MAX_SERVO_ABS else self.__class__._MAX_SERVO_ABS * (y / abs(y)))
         else:
-            self.__servo[1] = y
+            self.__servo[1] = int(y)
 
     def _write(self):
         data = struct.pack("<BBhh", self.__class__._HEADER, self.__class__._DEVID, self.__servo[0], self.__servo[1])
@@ -309,12 +321,12 @@ class Delta(Controller):
 
         for i, motor in enumerate(motors):
             if motor <= self.__class__._MAX_MT_POS and motor >= self.__class__._MIN_MT_POS:
-                self.__motors[i] = motor
+                self.__motors[i] = int(motor)
             else:
                 if motor >= self.__class__._MAX_MT_POS:
-                    self.__motors[i] = self.__class__._MAX_MT_POS
+                    self.__motors[i] = int(self.__class__._MAX_MT_POS)
                 else:
-                    self.__motors[i] = self.__class__._MIN_MT_POS
+                    self.__motors[i] = int(self.__class__._MIN_MT_POS)
 
     def _write(self):
         data = struct.pack("<BBBhhh", self.__class__._HEADER, self.__class__._DEVID, self.__magnet, *self.__motors)
@@ -341,11 +353,15 @@ class Stewart(Controller):
         self.position = [0] * 6
         self.imu = [0] * 3
 
+        if parse_version(self.board_info['Hardware Version']) <= parse_version('1.1.0'):
+            raise UnsupportedHardware("Stewart is only available on Acrome Controller hardware version 1.2.0 or later. Your version is {}".format(self.board_info['Hardware Version']))
+
+
     def __del__(self):
         super().__del__()
 
     def enable(self, en):
-        self.__en = en & 0x01
+        self.__en = (self.__en & ~0x01) | (en & 0x01)
 
     def set_motors(self, motors):
         if len(motors) != 6:
@@ -353,14 +369,15 @@ class Stewart(Controller):
 
         for i, motor in enumerate(motors):
             if motor != 0:
-                self.__motors[i] = motor if abs(motor) <= self.__class__._MAX_MT_ABS else self.__class__._MAX_MT_ABS * (motor / abs(motor))
+                self.__motors[i] = int(motor if abs(motor) <= self.__class__._MAX_MT_ABS else self.__class__._MAX_MT_ABS * (motor / abs(motor)))
             else:
-                self.__motors[i] = 0
+                self.__motors[i] = int(motor)
 
     def _write(self):
         data = struct.pack("<BBBhhhhhh", self.__class__._HEADER, self.__class__._DEVID, self.__en, *self.__motors)
         data += self._crc32(data)
         super()._writebus(data)
+        self.__en &= 0x01
 
     def _read(self):
         data = super()._readbus(self.__class__._RECEIVE_COUNT)
@@ -370,3 +387,107 @@ class Stewart(Controller):
                 self.imu = list(struct.unpack("<fff", data[14:26]))
                 return True
         return False
+
+class StewartEncoder(Controller):
+    _DEVID = 0xC0
+    _MAX_MT_ABS = 1000
+    _RECEIVE_COUNT = 30
+
+    def __init__(self, portname="/dev/serial0", baudrate=115200):
+        super().__init__(portname=portname, baudrate=baudrate)
+        self.__en = 0
+        self.__motors = [0] * 6
+        self.position = [0] * 6
+        self.imu = [0] * 3
+
+        if parse_version(self.board_info['Hardware Version']) <= parse_version('1.1.0'):
+            raise UnsupportedHardware("Stewart is only available on Acrome Controller hardware version 1.2.0 or later. Your version is {}".format(self.board_info['Hardware Version']))
+
+        if parse_version(self.board_info['Software Version']) < parse_version('1.5.0'):
+            raise UnsupportedFirmware("Stewart is only available on Acrome Controller software version 1.5.0 or later. Your version is {}".format(self.board_info['Software Version']))
+
+    def __del__(self):
+        super().__del__()
+
+    def enable(self, en):
+        self.__en = (self.__en & ~0x01) | (en & 0x01)
+    
+    def reset_encoder(self, motor_num=[1,2,3,4,5,6]):
+        for mt in motor_num:
+            if (mt <= 6 and mt >= 1):
+                self.__en |= 1 << (mt - 1 + 2)
+            else:
+                raise ValueError("Motor index can not be lower than 1 or higher than 6!")
+
+    def set_motors(self, motors):
+        if len(motors) != 6:
+            raise Exception("Argument motors must have length of 6")
+
+        for i, motor in enumerate(motors):
+            if motor != 0:
+                self.__motors[i] = int(motor if abs(motor) <= self.__class__._MAX_MT_ABS else self.__class__._MAX_MT_ABS * (motor / abs(motor)))
+            else:
+                self.__motors[i] = int(motor)
+
+    def _write(self):
+        data = struct.pack("<BBBhhhhhh", self.__class__._HEADER, self.__class__._DEVID, self.__en, *self.__motors)
+        data += self._crc32(data)
+        super()._writebus(data)
+        self.__en &= 0x01
+
+    def _read(self):
+        data = super()._readbus(self.__class__._RECEIVE_COUNT)
+        if data is not None:
+            if data[self.__class__._ID_INDEX] == self.__class__._DEVID:
+                self.position = list(struct.unpack("<HHHHHH", data[2:14]))
+                self.imu = list(struct.unpack("<fff", data[14:26]))
+                return True
+        return False
+
+class StewartEncoderHR(StewartEncoder):
+    _DEVID = 0xC1
+    _MAX_MT_ABS = 1000
+    _RECEIVE_COUNT = 42
+
+    def __init__(self, portname="/dev/serial0", baudrate=115200):
+        super().__init__(portname=portname, baudrate=baudrate)
+        self.__en = 0
+        self.__motors = [0] * 6
+        self.position = [0] * 6
+        self.imu = [0] * 3
+
+    def enable(self, en):
+        self.__en = (self.__en & ~0x01) | (en & 0x01)
+
+    def reset_encoder(self, motor_num=[1,2,3,4,5,6]):
+        for mt in motor_num:
+            if (mt <= 6 and mt >= 1):
+                self.__en |= 1 << (mt - 1 + 2)
+            else:
+                raise ValueError("Motor index can not be lower than 1 or higher than 6!")
+
+    def set_motors(self, motors):
+        if len(motors) != 6:
+            raise Exception("Argument motors must have length of 6")
+
+        for i, motor in enumerate(motors):
+            if motor != 0:
+                self.__motors[i] = int(motor if abs(motor) <= self.__class__._MAX_MT_ABS else self.__class__._MAX_MT_ABS * (motor / abs(motor)))
+            else:
+                self.__motors[i] = int(motor)
+
+
+    def _write(self):
+        data = struct.pack("<BBBhhhhhh", self.__class__._HEADER, self.__class__._DEVID, self.__en, *self.__motors)
+        data += self._crc32(data)
+        super()._writebus(data)
+        self.__en &= 0x01
+
+    def _read(self):
+        data = super()._readbus(self.__class__._RECEIVE_COUNT)
+        if data is not None:
+            if data[self.__class__._ID_INDEX] == self.__class__._DEVID:
+                self.position = list(struct.unpack("<IIIIII", data[2:26]))
+                self.imu = list(struct.unpack("<fff", data[26:38]))
+                return True
+        return False
\ No newline at end of file
